package javax.imageio.stream;

import java.io.IOException;
import sun.java2d.Disposer;
import com.sun.imageio.stream.StreamFinalizer;
import java.nio.file.Files;
import java.nio.file.attribute.FileAttribute;
import com.sun.imageio.stream.StreamCloser;
import sun.java2d.DisposerRecord;
import java.io.RandomAccessFile;
import java.io.File;
import java.io.InputStream;

public class FileCacheImageInputStream extends ImageInputStreamImpl
{
    private InputStream stream;
    private File cacheFile;
    private RandomAccessFile cache;
    private static final int BUFFER_LENGTH = 1024;
    private byte[] buf;
    private long length;
    private boolean foundEOF;
    private final Object disposerReferent;
    private final DisposerRecord disposerRecord;
    private final StreamCloser.CloseAction closeAction;
    
    public FileCacheImageInputStream(InputStream stream, File file) throws IOException {
        this.buf = new byte[1024];
        this.length = 0L;
        this.foundEOF = false;
        if (stream == null) {
            throw new IllegalArgumentException("stream == null!");
        }
        if (file != null && !file.isDirectory()) {
            throw new IllegalArgumentException("Not a directory!");
        }
        this.stream = stream;
        if (file == null) {
            this.cacheFile = Files.createTempFile("imageio", ".tmp", new FileAttribute[0]).toFile();
        }
        else {
            this.cacheFile = Files.createTempFile(file.toPath(), "imageio", ".tmp", new FileAttribute[0]).toFile();
        }
        this.cache = new RandomAccessFile(this.cacheFile, "rw");
        StreamCloser.addToQueue(this.closeAction = StreamCloser.createCloseAction(this));
        this.disposerRecord = new FileCacheImageInputStream.StreamDisposerRecord(this.cacheFile, this.cache);
        if (this.getClass() == FileCacheImageInputStream.class) {
            Disposer.addRecord(this.disposerReferent = new Object(), this.disposerRecord);
        }
        else {
            this.disposerReferent = new StreamFinalizer(this);
        }
        return;
    }
    
    private long readUntil(long n) throws IOException {
        if (n < this.length) {
            return n;
        }
        if (this.foundEOF) {
            return this.length;
        }
        long a = n - this.length;
        this.cache.seek(this.length);
        while (a > 0L) {
            int read = this.stream.read(this.buf, 0, (int)Math.min(a, 1024L));
            if (read == -1) {
                this.foundEOF = true;
                return this.length;
            }
            this.cache.write(this.buf, 0, read);
            a -= (long)read;
            this.length += (long)read;
        }
        return n;
    }
    
    @Override
    public int read() throws IOException {
        this.checkClosed();
        this.bitOffset = 0;
        long n = this.streamPos + 1L;
        if (this.readUntil(n) >= n) {
            this.cache.seek(this.streamPos++);
            return this.cache.read();
        }
        return -1;
    }
    
    @Override
    public int read(byte[] b, int off, int len) throws IOException {
        this.checkClosed();
        if (b == null) {
            throw new NullPointerException("b == null!");
        }
        if (off < 0 || len < 0 || off + len > b.length || off + len < 0) {
            throw new IndexOutOfBoundsException("off < 0 || len < 0 || off+len > b.length || off+len < 0!");
        }
        this.bitOffset = 0;
        if (len == 0) {
            return 0;
        }
        len = (int)Math.min((long)len, this.readUntil(this.streamPos + (long)len) - this.streamPos);
        if (len > 0) {
            this.cache.seek(this.streamPos);
            this.cache.readFully(b, off, len);
            this.streamPos += (long)len;
            return len;
        }
        return -1;
    }
    
    @Override
    public boolean isCached() {
        return true;
    }
    
    @Override
    public boolean isCachedFile() {
        return true;
    }
    
    @Override
    public boolean isCachedMemory() {
        return false;
    }
    
    @Override
    public void close() throws IOException {
        super.close();
        this.disposerRecord.dispose();
        this.stream = null;
        this.cache = null;
        this.cacheFile = null;
        StreamCloser.removeFromQueue(this.closeAction);
    }
    
    @Override
    protected void finalize() throws Throwable {
    }
    
    private static class StreamDisposerRecord implements DisposerRecord
    {
        private File cacheFile;
        private RandomAccessFile cache;
        
        public StreamDisposerRecord(File cacheFile, RandomAccessFile cache) {
            this.cacheFile = cacheFile;
            this.cache = cache;
        }
        
        @Override
        public synchronized void dispose() {
            if (this.cache != null) {
                try {
                    this.cache.close();
                }
                catch (IOException ex) {}
                finally {
                    this.cache = null;
                }
            }
            if (this.cacheFile != null) {
                this.cacheFile.delete();
                this.cacheFile = null;
            }
        }
    }
}
